import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import minimize

def DataModel(x):
    return 3.25*x+0.13

def GetNoisyData(N):
    x = np.sort( np.random.rand(N) )
    noise = 0.5 * (np.random.rand(N) - 0.5)
    y = DataModel(x) + noise
    return x, y

def LeastSquaresResidual(w, Amat, y_tilde):
    res = np.matmul(Amat, w) - y_tilde
    return np.dot(res, res)

def RegularizedLeastSquaresResidual(w, Amat, y_tilde):
    n = np.arange(0, w.shape[0], 1)
    reg = np.sum(n * w**2)
    return LeastSquaresResidual(w, Amat, y_tilde) + 0.05 * reg

N = 10
x, y_tilde = GetNoisyData(N)

"""=== Linear Least Squares ==="""
Amat1 = np.vstack( (x, np.ones(x.shape)) ).T

# using pseudo-inverse
wRef = np.matmul(np.linalg.pinv(Amat1), y_tilde)

# using minimization
wInit = 2 * (np.random.rand(Amat1.shape[1]) - 0.5)
w = minimize(LeastSquaresResidual, wInit, args=(Amat1, y_tilde), tol=1E-6)

"""=== Exact Interpolation ==="""
pdeg = 9
p = np.linspace(pdeg, 0, pdeg+1)
Amatn = x.reshape(-1,1)**p

#solve linear system
w_exact = np.linalg.solve(Amatn, y_tilde)

# evaluate the polynomial
x_points = np.linspace(0,1,501)
polynomial = np.polyval(w_exact, x_points)

"""=== Underdetermined setting ==="""
pdeg = 20
p = np.linspace(pdeg, 0, pdeg+1)
Amatn = x.reshape(-1,1)**p

wInit = 2 * (np.random.rand(pdeg+1) -0.5)
w_under = minimize(LeastSquaresResidual, wInit, args=(Amatn, y_tilde))

# evaluate the polynomial
polynomial_under = np.polyval(w_under.x, x_points)

"""=== Underdetermined setting + Regularization ==="""
w_under_reg = minimize(RegularizedLeastSquaresResidual, wInit, args=(Amatn, y_tilde))

# evaluate the polynomial
polynomial_under_reg = np.polyval(w_under_reg.x, x_points)

"""=== PLOTTING ==="""
def set_plot_properties(ax):
    ax.spines["top"].set_visible(False)
    ax.spines["right"].set_visible(False)
    ax.set(xlim=(0., 1.1))

# plot data points
fig1, ax1 = plt.subplots()
ax1.scatter(x, y_tilde, color="gray")
set_plot_properties(ax1)
plt.title("Data Points")

# Linear Regression
fig2, ax2 = plt.subplots()
ax2.scatter(x, y_tilde, color="gray")
ax2.plot(x_points, w.x[1] + w.x[0]*x_points, "--")
set_plot_properties(ax2)
plt.title("Linear Regression")

# Exact Interpolation
fig3, ax3 = plt.subplots()
ax3.scatter(x, y_tilde, color="gray")
ax3.plot(x_points, polynomial, "--")
set_plot_properties(ax3)
plt.title("Exact Polynomial Interpolation")

# Underdetermined Setting
fig4, ax4 = plt.subplots()
ax4.scatter(x, y_tilde, color="gray")
ax4.plot(x_points, polynomial_under, "--")
set_plot_properties(ax4)
plt.title("Underdetermined Polynomial")

# Underdetermined Setting Coefficients
fig5, ax5 = plt.subplots()
for i in range(w_under.x.shape[0]):
    ax5.scatter(i, w_under.x[i], color="tab:blue", s=80)
    ax5.plot([i, i], [0, w_under.x[i]], color="tab:blue", lw=2)
set_plot_properties(ax5)
ax5.spines["bottom"].set_position("zero")
ax5.set(xlim=(0, w_under.x.shape[0]))
xticks = np.arange(0, np.floor(w_under.x.shape[0]/4))
ax5.set_xticks(4 * xticks[xticks != 0.])
ax5.set_xlabel(r"$a_n$")
plt.title("Coefficients of Underdetermined Polynomial")

# Underdetermined Setting + Regularization
fig6, ax6 = plt.subplots()
ax6.scatter(x, y_tilde, color="gray")
ax6.plot(x_points, polynomial_under_reg, "--")
set_plot_properties(ax6)
plt.title("Underdetermined Polynomial + Regularization")

# Underdetermined Setting Coefficients
fig7, ax7 = plt.subplots()
for i in range(w_under_reg.x.shape[0]):
    ax7.scatter(i, w_under_reg.x[i], color="tab:blue", s=80)
    ax7.plot([i, i], [0, w_under_reg.x[i]], color="tab:blue", lw=2)
set_plot_properties(ax7)
ax7.spines["bottom"].set_position("zero")
ax7.set(xlim=(0, w_under_reg.x.shape[0]))
xticks = np.arange(0, np.floor(w_under_reg.x.shape[0]/4))
ax7.set_xticks(4 * xticks[xticks != 0.])
ax7.set_xlabel(r"$a_n$")
plt.title("Coefficients of Underdetermined Polynomial + Regularization")

plt.show()
